@@ -308,6 +308,7 @@ static struct l_dhcp_lease *add_lease(struct l_dhcp_server *server,
 	if (ret != 0)
 		return NULL;
 
+	l_free(lease->dns);
 	l_free(lease->client_id);
 	memset(lease, 0, sizeof(*lease));
 
@@ -316,6 +317,13 @@ static struct l_dhcp_lease *add_lease(struct l_dhcp_server *server,
 	lease->subnet_mask = server->netmask;
 	lease->router = server->gateway;
 
+	if (server->dns_list) {
+		unsigned int i;
+
+		for (i = 0; server->dns_list[i]; i++);
+		lease->dns = l_memdup(server->dns_list, (i + 1) * 4);
+	}
+
 	if (client_id)
 		lease->client_id = l_memdup(client_id, client_id[0] + 1);
 
@@ -1025,10 +1033,6 @@ LIB_EXPORT bool l_dhcp_server_start(struct l_dhcp_server *server)
 		server->address = ia.s_addr;
 	}
 
-	/* Assign a default gateway if not already set */
-	if (!server->gateway)
-		server->gateway = server->address;
-
 	/* Assign a default netmask if not already */
 	if (!server->netmask) {
 		if (inet_pton(AF_INET,"255.255.255.0", &ia) != 1)
